@@ -28,9 +28,11 @@
 #include "config.h"
 #endif
 
+#include <xorg-server.h>
+#include <xorgVersion.h>
+
 #include <xf86.h>
 #include <xf86Parser.h>
-#include <xorgVersion.h>
 
 #if XORG_VERSION_CURRENT < XORG_VERSION_NUMERIC(1,6,99,0,0)
 #include <xf86Resources.h>
@@ -207,14 +209,14 @@ static const SymTabRec intel_chipsets[] = {
 #define NUM_CHIPSETS (sizeof(intel_chipsets) / sizeof(intel_chipsets[0]))
 
 static const struct pci_id_match intel_device_match[] = {
-#if !KMS_ONLY
+#if UMS
 	INTEL_VGA_DEVICE(PCI_CHIP_I810, &intel_i81x_info),
 	INTEL_VGA_DEVICE(PCI_CHIP_I810_DC100, &intel_i81x_info),
 	INTEL_VGA_DEVICE(PCI_CHIP_I810_E, &intel_i81x_info),
 	INTEL_VGA_DEVICE(PCI_CHIP_I815, &intel_i81x_info),
 #endif
 
-#if !UMS_ONLY
+#if KMS
 	INTEL_I830_IDS(&intel_i830_info),
 	INTEL_I845G_IDS(&intel_i830_info),
 	INTEL_I85X_IDS(&intel_i855_info),
@@ -256,23 +258,30 @@ static const struct pci_id_match intel_device_match[] = {
 };
 
 void
-intel_detect_chipset(ScrnInfoPtr scrn,
-		     EntityInfoPtr ent,
-		     struct pci_device *pci)
+intel_detect_chipset(ScrnInfoPtr scrn, EntityInfoPtr ent)
 {
 	MessageType from = X_PROBED;
 	const char *name = NULL;
+	int devid;
 	int i;
 
 	if (ent->device->chipID >= 0) {
 		xf86DrvMsg(scrn->scrnIndex, from = X_CONFIG,
 			   "ChipID override: 0x%04X\n",
 			   ent->device->chipID);
-		pci->device_id = ent->device->chipID;
+		devid = ent->device->chipID;
+	} else {
+		struct pci_device *pci;
+
+		pci = xf86GetPciInfoForEntity(ent->index);
+		if (pci != NULL)
+			devid = pci->device_id;
+		else
+			devid = intel_get_device_id(scrn);
 	}
 
 	for (i = 0; intel_chipsets[i].name != NULL; i++) {
-		if (pci->device_id == intel_chipsets[i].token) {
+		if (devid == intel_chipsets[i].token) {
 			name = intel_chipsets[i].name;
 			break;
 		}
@@ -281,20 +290,21 @@ intel_detect_chipset(ScrnInfoPtr scrn,
 		int gen = 0;
 
 		for (i = 0; intel_device_match[i].device_id != 0; i++) {
-			if (pci->device_id == intel_device_match[i].device_id) {
+			if (devid == intel_device_match[i].device_id) {
 				const struct intel_device_info *info = (void *)intel_device_match[i].match_data;
 				gen = info->gen >> 3;
 				break;
 			}
 		}
 
-		if(gen) {
+		if (gen) {
 			xf86DrvMsg(scrn->scrnIndex, from,
 				   "gen%d engineering sample\n", gen);
 		} else {
 			xf86DrvMsg(scrn->scrnIndex, X_WARNING,
 				   "Unknown chipset\n");
 		}
+
 		name = "unknown";
 	} else {
 		xf86DrvMsg(scrn->scrnIndex, from,
@@ -377,11 +387,15 @@ static Bool intel_driver_func(ScrnInfoPtr pScrn,
 	switch (op) {
 	case GET_REQUIRED_HW_INTERFACES:
 		flag = (CARD32*)ptr;
-#ifdef KMS_ONLY
 		(*flag) = 0;
-#else
+#if UMS
 		(*flag) = HW_IO | HW_MMIO;
 #endif
+#ifdef HW_SKIP_CONSOLE
+		if (hosted())
+			(*flag) = HW_SKIP_CONSOLE;
+#endif
+
 		return TRUE;
 	default:
 		/* Unknown or deprecated function */
@@ -389,7 +403,7 @@ static Bool intel_driver_func(ScrnInfoPtr pScrn,
 	}
 }
 
-#if !UMS_ONLY
+#if KMS
 extern XF86ConfigPtr xf86configptr;
 
 static XF86ConfDevicePtr
@@ -410,6 +424,9 @@ static enum accel_method { UXA, SNA } get_accel_method(void)
 	enum accel_method accel_method = DEFAULT_ACCEL_METHOD;
 	XF86ConfDevicePtr dev;
 
+	if (hosted())
+		return SNA;
+
 	dev = _xf86findDriver("intel", xf86configptr->conf_device_lst);
 	if (dev && dev->dev_option_lst) {
 		const char *s;
@@ -437,6 +454,22 @@ intel_scrn_create(DriverPtr		driver,
 {
 	ScrnInfoPtr scrn;
 
+	if (match_data == 0) {
+		int devid = intel_entity_get_devid(entity_num), i;
+		if (devid == 0)
+			return FALSE;
+
+		for (i = 0; intel_device_match[i].device_id != 0; i++) {
+			if (devid == intel_device_match[i].device_id) {
+				match_data = (intptr_t)&intel_device_match[i];
+				break;
+			}
+		}
+
+		if (match_data == 0)
+			return FALSE;
+	}
+
 	scrn = xf86AllocateScreen(driver, flags);
 	if (scrn == NULL)
 		return FALSE;
@@ -451,12 +484,12 @@ intel_scrn_create(DriverPtr		driver,
 		xf86SetEntityShared(entity_num);
 	xf86AddEntityToScreen(scrn, entity_num);
 
-#if !KMS_ONLY
+#if UMS
 	if ((unsigned)((struct intel_device_info *)match_data)->gen < 020)
 		return lg_i810_init(scrn);
 #endif
 
-#if !UMS_ONLY
+#if KMS
 	switch (get_accel_method()) {
 #if USE_SNA
 	case SNA: return sna_init_scrn(scrn, entity_num);
@@ -485,18 +518,19 @@ static Bool intel_pci_probe(DriverPtr		driver,
 			    intptr_t		match_data)
 {
 	if (intel_open_device(entity_num, pci, NULL) == -1) {
-#if KMS_ONLY
-		return FALSE;
-#else
+#if UMS
 		switch (pci->device_id) {
 		case PCI_CHIP_I810:
 		case PCI_CHIP_I810_DC100:
 		case PCI_CHIP_I810_E:
 		case PCI_CHIP_I815:
-			break;
+			if (!hosted())
+				break;
 		default:
 			return FALSE;
 		}
+#else
+		return FALSE;
 #endif
 	}
 
@@ -512,11 +546,7 @@ intel_platform_probe(DriverPtr driver,
 {
 	unsigned scrn_flags = 0;
 
-	if (!dev->pdev)
-		return FALSE;
-
-	if (intel_open_device(entity_num, dev->pdev,
-			      xf86_get_platform_device_attrib(dev, ODEV_ATTRIB_PATH)) == -1)
+	if (intel_open_device(entity_num, dev->pdev, dev) == -1)
 		return FALSE;
 
 	/* Allow ourselves to act as a slaved output if not primary */
@@ -554,7 +584,7 @@ static const OptionInfoRec *
 intel_available_options(int chipid, int busid)
 {
 	switch (chipid) {
-#if !KMS_ONLY
+#if UMS
 	case PCI_CHIP_I810:
 	case PCI_CHIP_I810_DC100:
 	case PCI_CHIP_I810_E:
