@@ -151,7 +151,7 @@ struct sna_composite_op {
 struct sna_opacity_box {
 	BoxRec box;
 	float alpha;
-} __packed__;
+} tightly_packed;
 
 struct sna_composite_spans_op {
 	struct sna_composite_op base;
@@ -255,8 +255,11 @@ struct sna_render {
 			   const BoxRec *box, int n);
 	bool (*fill)(struct sna *sna, uint8_t alu,
 		     PixmapPtr dst, struct kgem_bo *dst_bo,
-		     uint32_t color,
+		     uint32_t color, unsigned flags,
 		     struct sna_fill_op *tmp);
+#define FILL_BOXES 0x1
+#define FILL_POINTS 0x2
+#define FILL_SPANS 0x4
 	bool (*fill_one)(struct sna *sna, PixmapPtr dst, struct kgem_bo *dst_bo,
 			 uint32_t color,
 			 int16_t x1, int16_t y1, int16_t x2, int16_t y2,
@@ -287,7 +290,7 @@ struct sna_render {
 	struct sna_solid_cache {
 		struct kgem_bo *cache_bo;
 		struct kgem_bo *bo[1024];
-		uint32_t color[1025];
+		uint32_t color[1024];
 		int last;
 		int size;
 		int dirty;
@@ -391,10 +394,10 @@ struct gen5_render_state {
 	int ve_id;
 	uint32_t drawrect_offset;
 	uint32_t drawrect_limit;
+	uint32_t last_pipelined_pointers;
 	uint16_t last_primitive;
 	int16_t floats_per_vertex;
 	uint16_t surface_table;
-	uint16_t last_pipelined_pointers;
 
 	bool needs_invariant;
 };
@@ -421,6 +424,7 @@ enum {
 };
 
 struct gen6_render_state {
+	unsigned gt;
 	const struct gt_info *info;
 	struct kgem_bo *general_bo;
 
@@ -470,6 +474,7 @@ enum {
 };
 
 struct gen7_render_state {
+	unsigned gt;
 	const struct gt_info *info;
 	struct kgem_bo *general_bo;
 
@@ -775,4 +780,11 @@ static inline bool sna_vertex_wait__locked(struct sna_render *r)
 	return was_active;
 }
 
+#define alphaless(format) PICT_FORMAT(PICT_FORMAT_BPP(format),		\
+				      PICT_FORMAT_TYPE(format),		\
+				      0,				\
+				      PICT_FORMAT_R(format),		\
+				      PICT_FORMAT_G(format),		\
+				      PICT_FORMAT_B(format))
+
 #endif /* SNA_RENDER_H */
