@@ -55,6 +55,7 @@
 #include <assert.h>
 #include <errno.h>
 
+#include "xorg-server.h"
 #include "xf86.h"
 #include "xf86_OSproc.h"
 #include "compiler.h"
@@ -1017,9 +1018,14 @@ static void intel_box_intersect(BoxPtr dest, BoxPtr a, BoxPtr b)
 {
 	dest->x1 = a->x1 > b->x1 ? a->x1 : b->x1;
 	dest->x2 = a->x2 < b->x2 ? a->x2 : b->x2;
+	if (dest->x1 >= dest->x2) {
+		dest->x1 = dest->x2 = dest->y1 = dest->y2 = 0;
+		return;
+	}
+
 	dest->y1 = a->y1 > b->y1 ? a->y1 : b->y1;
 	dest->y2 = a->y2 < b->y2 ? a->y2 : b->y2;
-	if (dest->x1 >= dest->x2 || dest->y1 >= dest->y2)
+	if (dest->y1 >= dest->y2)
 		dest->x1 = dest->x2 = dest->y1 = dest->y2 = 0;
 }
 
